"""
Resoto common library
~~~~~~~~~~~~~~~~~~~~~~~~~~
Common Resoto functions.
:copyright: © 2023 Some Engineering Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

__title__ = "resotolib"
__description__ = "Resoto common library."
__author__ = "Some Engineering Inc."
__license__ = "Apache 2.0"
__copyright__ = "Copyright © 2023 Some Engineering Inc."
__version__ = "3.5.0"
